#include <SFML/Graphics.hpp>
#include <list>
#include <set>
#include <cmath> // atan2(), sin(), cost()

#define DEBUG

/* Sources :
 * https://ncase.me/sight-and-light/
 * https://stackoverflow.com/questions/563198/how-do-you-detect-where-two-line-segments-intersect
 */

 // TODO : https://docs.gimp.org/2.10/en/gimp-filter-normal-map.html

// Returns the intersection between segments [p0,p1] and [p2,p3], or throws if no collision.
sf::Vector2f getIntersection(sf::Vector2f p0, sf::Vector2f p1, sf::Vector2f p2, sf::Vector2f p3)
	{
	sf::Vector2f s1 = p1 - p0;
	sf::Vector2f s2 = p3 - p2;

	float s = (-s1.y * (p0.x - p2.x) + s1.x * (p0.y - p2.y)) / (-s2.x * s1.y + s1.x * s2.y);
	float t = ( s2.x * (p0.y - p2.y) - s2.y * (p0.x - p2.x)) / (-s2.x * s1.y + s1.x * s2.y);

	if (0.f <= s and s <= 1.f and 0.f <= t and t <= 1.f) {
		// Collision detected
		sf::Vector2f i = p0 + s1*t;
		/*
		i_x = p0_x + (t * s1_x);
		i_y = p0_y + (t * s1_y);
		*/
		return i;
	}
	throw "plop"; // std::out_of_range; // No collision
}

// Returns the intersection between the ray cast from the origin at a given angle and the segment [p2,p3], or throws if no collision.
sf::Vector2f getIntersection(sf::Vector2f origin, float angle, sf::Vector2f p2, sf::Vector2f p3) {
	sf::Vector2f dir(std::cos(angle), std::sin(angle));
	sf::Vector2f p1 = origin + dir*1000.f;
	return(getIntersection(origin, p1, p2, p3));
}

// Returns the angle of the ray cast from origin toward target.
float getAngle(sf::Vector2f origin, sf::Vector2f target) {
	return(std::atan2(target.y - origin.y, target.x - origin.x));
}

// Returns the distance between p0 and p1.
float getDistance(sf::Vector2f p0, sf::Vector2f p1) {
	float a = p0.x - p1.x;
	float b = p0.y - p1.y;
	return(std::sqrt(a*a+b*b));
}

// Returns the closest intersection between a ray (cast from origin at given angle) and a list of obstacles.
sf::Vector2f getClosestHit(sf::Vector2f origin, float angle, const std::list<std::pair<sf::Vector2f,sf::Vector2f>>& obstacles) {
	std::list<sf::Vector2f> hits;
	sf::Vector2f direction(std::cos(angle), std::sin(angle));
	sf::Vector2f farPoint = origin + direction*1000.f;

	for(const std::pair<sf::Vector2f,sf::Vector2f>& obstacle : obstacles) {
		try {
			hits.push_back(getIntersection(origin, farPoint, obstacle.first, obstacle.second));
		} catch(...) {
		}
	}

	if(hits.empty()) {
		throw "plop"; // std::out_of_range; // No collision
	}

	sf::Vector2f closestHit = hits.front();
	float closestDistance = getDistance(origin, closestHit);
	for(sf::Vector2f hit : hits) {
		float dist = getDistance(origin, hit);
		if(dist < closestDistance) {
			closestDistance = dist;
			closestHit = hit;
		}
	}
	return(closestHit);
}

class LightSource : public sf::Drawable {
public:
	LightSource(sf::Vector2f position, sf::Color color = sf::Color::White) :
		position { position },
		color { color },
		area { sf::TriangleFan }
	{
	}

	sf::Color getColor() const {
		return(this->color);
	}

	void setColor(sf::Color newColor) {
		this->color = newColor;
	}

	void setPosition(sf::Vector2f newPos, const std::list<std::pair<sf::Vector2f,sf::Vector2f>>& obstacles, const std::list<sf::Vector2f>& targets) {
		this->position = newPos;
		this->update(obstacles, targets);
	}

	void update(const std::list<std::pair<sf::Vector2f,sf::Vector2f>>& obstacles, const std::list<sf::Vector2f>& targets) {
		/* Raycast. */

		std::map<float,sf::Vector2f> hits; // Raycast's hits, stored by their angle, so they can be retrieved in order.
		for(sf::Vector2f target : targets) {
			float directAngle = getAngle(this->position, target);
			for(auto angle : {directAngle - 0.00001f, directAngle, directAngle + 0.00001f}) {
				try {
					sf::Vector2f closestHit = getClosestHit(this->position, angle, obstacles);
					hits[angle] = closestHit;
				} catch(...) { }
			}
		}

		/* Define the lit area as a triangle-fan having the light source as center and each clockwise hit defining the fanned triangles. */

		this->area.clear();
		this->area.append(sf::Vertex(this->position));
		for(const auto& hit : hits) {
			this->area.append(sf::Vertex(hit.second));
		}
		this->area.append(sf::Vertex(hits.begin()->second));
	}

	virtual void draw(sf::RenderTarget& target, sf::RenderStates states) const override {
		target.draw(this->area, states);
	}

	void preDraw(sf::Shader& shader) const {
		shader.setUniform("lightpos", this->position);
		shader.setUniform("lightColor", sf::Glsl::Vec4(this->color));
	}
private:
	sf::Vector2f position;
	sf::Color color;
	sf::VertexArray area;
};

int main() {
	sf::RenderWindow window(sf::VideoMode(800, 600), "Shader Crashtest");

	sf::Texture infoTexture;
	if(not infoTexture.loadFromFile("info.png")) {
		// TODO : error.
	}
	sf::Sprite infoSprite(infoTexture);

	/* Define shader. */

	sf::Shader shader;
	if(not shader.loadFromFile("LightFS.txt", sf::Shader::Fragment)) {
		// TODO : error.
	}

	// shader.setUniform("lightpos", sf::Vector2f(100.0, 50.0)); // Per-light parameter.
	// shader.setUniform("lightColor", sf::Vector3f(128.0, 0.0, 128.0)); // Per-light parameter.
	shader.setUniform("screenHeight", 600.f); // Global parameter.
	shader.setUniform("lightAttenuation", sf::Vector3f(1.0, 0.01, 0.001)); // Technically per-light parameter, but used globally for now.
	// shader.setUniform("radius", float(50.f)); // Unused parameter.

	std::vector<sf::Color> colors = {
		sf::Color::Red,
		sf::Color::White,
		sf::Color::Yellow,
		sf::Color::Green,
		sf::Color::Cyan,
		sf::Color::Blue,
		sf::Color::Magenta
	};
	int colorIndex = colors.size()-1;

	std::list<LightSource> lightSources;
	lightSources.push_back(LightSource(sf::Vector2f(0.0, 0.0), colors[colorIndex]));

	/* Define obstacles. */

	std::list<std::pair<sf::Vector2f,sf::Vector2f>> obstacles;

	// Screen's bounds.
	obstacles.push_back({sf::Vector2f(0, 0), sf::Vector2f(800, 0)});
	obstacles.push_back({sf::Vector2f(800, 0), sf::Vector2f(800, 600)});
	obstacles.push_back({sf::Vector2f(800, 600), sf::Vector2f(0, 600)});
	obstacles.push_back({sf::Vector2f(0, 600), sf::Vector2f(0, 0)});

	// Triangle.
	obstacles.push_back({sf::Vector2f(100, 100), sf::Vector2f(300, 100)});
	obstacles.push_back({sf::Vector2f(300, 100), sf::Vector2f(300, 500)});
	obstacles.push_back({sf::Vector2f(300, 500), sf::Vector2f(100, 100)});

	// Square.
	obstacles.push_back({sf::Vector2f(500, 200), sf::Vector2f(500, 300)});
	obstacles.push_back({sf::Vector2f(500, 300), sf::Vector2f(600, 300)});
	obstacles.push_back({sf::Vector2f(600, 300), sf::Vector2f(600, 200)});
	obstacles.push_back({sf::Vector2f(600, 200), sf::Vector2f(500, 200)});

	// Walls.
	obstacles.push_back({sf::Vector2f(400, 400), sf::Vector2f(500, 500)});
	obstacles.push_back({sf::Vector2f(500, 500), sf::Vector2f(500, 400)});
	obstacles.push_back({sf::Vector2f(500, 400), sf::Vector2f(600, 500)});

	/* Define drawable lines of obstacles. */

	sf::VertexArray obstaclesLines(sf::Lines);
	for(std::pair<sf::Vector2f,sf::Vector2f> line : obstacles) {
		obstaclesLines.append(line.first);
		obstaclesLines.append(line.second);
	}

	/* Prepare a set of raycast targets (a set makes them unique, to avoid duplicates). */

	struct comparator { bool operator() (const sf::Vector2f& lhs, const sf::Vector2f& rhs) const {
		return((lhs.x == rhs.x) ? (lhs.y < rhs.y) : (lhs.x < rhs.x));
	} };
	std::set<sf::Vector2f, comparator> targetsSet;
	for(std::pair<sf::Vector2f,sf::Vector2f> obstacle : obstacles) {
		targetsSet.insert(obstacle.first);
		targetsSet.insert(obstacle.second);
	}
	std::list<sf::Vector2f> targets { targetsSet.begin(), targetsSet.end() };

	while (window.isOpen()) {

		/* Handle events. */

		sf::Event event;
		while (window.pollEvent(event)) {
			if (event.type == sf::Event::Closed or (event.type == sf::Event::KeyPressed and event.key.code == sf::Keyboard::Escape)) {
				window.close();
			} else if(event.type == sf::Event::MouseButtonPressed and event.mouseButton.button == sf::Mouse::Left) {
				LightSource newLightSource(sf::Vector2f(event.mouseButton.x, event.mouseButton.y), lightSources.front().getColor());
				newLightSource.update(obstacles, targets);
				lightSources.push_back(newLightSource);
			} else if(event.type == sf::Event::MouseButtonPressed and event.mouseButton.button == sf::Mouse::Right) {
				if(lightSources.size() > 1) {
					lightSources.pop_back();
				}
			} else if(event.type == sf::Event::MouseWheelScrolled) {
				colorIndex = (colorIndex + int(event.mouseWheelScroll.delta) + colors.size()) % colors.size();
				lightSources.front().setColor(colors[colorIndex]);
			}
		}

		/* Draw. */

		lightSources.front().setPosition(sf::Vector2f(sf::Mouse::getPosition(window)), obstacles, targets);

		window.clear();
		for(const LightSource& lightSource : lightSources) {
			lightSource.preDraw(shader);
			window.draw(lightSource, &shader);
		}
		window.draw(obstaclesLines);
		window.draw(infoSprite);
		window.display();
	}

	return 0;
}
